local level_id = Global.level_data and Global.level_data.level_id or ""
local difficulty = Global.game_settings and Global.game_settings.difficulty or "normal"

local vec3_100 = Vector3(0,100,0)
local vec3_200 = Vector3(0,200,0)

local allowed_to_search = {
	["security"] = true,
	["security_patrol"] = true
}

-- During first search phase, the added range will be 100m from your initial position
-- Second search phase will be 200m
function GroupAIStateBase:_getSearchRadius()
	if self._search_phase == "alert" then
		return vec3_100
	end
	
	return vec3_200
end

function GroupAIStateBase:_get_search_position(unit)
	local player = managers.player:player_unit()
	if not alive(player) then
		return
	end
	if not alive(unit) then
		return
	end
	
	local myPos = unit:position()
	
	local function getPosition()
		if self._search_phase == "alert" then
			return myPos + self:_getSearchRadius()
		end
		
		return player:position() + self:_getSearchRadius()
	end
	
	local pos = getPosition()
	local nav_seg = managers.navigation:get_nav_seg_from_pos(pos)
	local action_duration = self._search_phase == "alert" and math.random(6,8) or math.random(3,4)
	
	return pos, nav_seg, action_duration
end

local investigating_color = Color(255,255,255)

function GroupAIStateBase:_next_search_pos(unit)
	-- If we are not the server, abort
	if not Network:is_server() then
		return
	end
	
	-- If navigation is not ready, abort
	if not managers.navigation:is_data_ready() then
		return
	end
	
	local obs_key = unit:key()
	local susp_data = self._suspicion_hud_data
	local obs_susp_data = susp_data[obs_key]
	local icon_color = self._search_phase == "alert" and investigating_color or tweak_data.hud.suspicion_color
	
	if not obs_susp_data then
		local icon_id = "susp1" .. tostring(obs_key)
		local icon_pos = self._create_hud_suspicion_icon(obs_key, unit, "wp_suspicious", icon_color, icon_id)
		obs_susp_data = {
			u_observer = unit,
			icon_id = icon_id,
			icon_pos = icon_pos
		}
		susp_data[obs_key] = obs_susp_data
		
		managers.hud:change_waypoint_icon(obs_susp_data.icon_id, "wp_suspicious")
		managers.hud:change_waypoint_arrow_color(obs_susp_data.icon_id, icon_color)
		
		if Network:is_server() and managers.network:session() then
			managers.network:session():send_to_peers_synched("suspicion_hud", unit, 1)
		end
	end
	
	if self._search_phase == "alert2" or self._search_phase == "lockdown" then
		if unit:movement():stance_code() == 1 then
			unit:movement():_change_stance(2, true)
		end
	end
	
	-- Guards will either search your last known position or search the current area
	local pos, nav_seg, action_duration = self:_get_search_position(unit)
	
	if not self:is_nav_seg_safe(nav_seg) then
		return
	end
	
	local area = self:get_area_from_nav_seg_id(nav_seg)
	if not area or area and area.disabled then
		return
	end
	
	local function getWalkStyle()
		if self._search_phase == "alert2" or self._search_phase == "lockdown" then
			return "run"
		end
		
		return "walk"
	end
	
	local search_so = {
		type = "search",
		attitude = "engage",
		pose = "stand",
		haste = getWalkStyle(),
		scan = true,
		forced = true,
		interrupt_dis = -1,
		pos = pos,
		nav_seg = nav_seg,
		area = area,
		action_duration = action_duration
	}
	
	local old_so = unit:brain():objective()
	if (old_so and old_so.path_style == "destination") then
		-- Currently moving towards patrol point, i will search the area later
		old_so.followup_objective = {
			type = "free",
			scan = true,
			forced = true,
			attitude = "avoid"
		}
	else
		-- I have no objective, so i will search now
		unit:brain():set_objective(search_so)
	end
end

function GroupAIStateBase:_upd_search_mode()
	for u_key,u_data in pairs(managers.enemy:all_enemies()) do
		if alive(u_data.unit) and u_data.unit:movement():cool() and allowed_to_search[u_data.unit:base():char_tweak().access] then
			self:_next_search_pos(u_data.unit)
		end
	end
end

function GroupAIStateBase:_clear_brain()
	for u_key,u_data in pairs(managers.enemy:all_enemies()) do
		if alive(u_data.unit) and u_data.unit:movement():cool() and u_data.unit:brain():objective() and allowed_to_search[u_data.unit:base():char_tweak().access] then
			u_data.unit:brain():set_objective()
		end
	end
end

function GroupAIStateBase:is_search_mode_active()
	return self._is_search_mode_active
end

function GroupAIStateBase:get_search_phase()
	return self._search_phase
end

function GroupAIStateBase:is_difficulty_greater()
	if difficulty == "overkill_145" or difficulty == "easy_wish" or difficulty == "overkill_290" or difficulty == "sm_wish" then
		return true
	end
	
	return false
end

function GroupAIStateBase:set_search_mode(state, reason)
	if not state or type(state) ~= "boolean" then return end
	if not self:whisper_mode() then return end
	
	if self:get_search_phase() == "lockdown" then
		if Global.game_settings.single_player then
			self:on_police_called("empty")
			managers.hud:show_hint({text = "ALARM TRIPPED: SECURITY WAS TOO SUSPICIOUS"})
		else
			self:on_police_called("sys_police_alerted")
		end
		return
	end
	
	if self._search_phase == "calm" then
		self._search_phase = "alert"
	elseif self._search_phase == "alert" then
		self._search_phase = "alert2"
	elseif self._search_phase == "alert2" then
		self._search_phase = "lockdown"
	end
	
	self:_clear_brain()
	self._is_search_mode_active = state
	
	local hud_text = "SEARCH MODE TRIGGERED"
	
	if not reason then
		if self._search_phase == "alert" then
			reason = "SECURITY IS NOW SUSPICIOUS AND WILL LOOK FOR YOU"
		elseif self._search_phase == "alert2" then
			reason = "SECURITY IS NOW ON ALERT"
		elseif self._search_phase == "lockdown" then
			reason = "SECURITY HAS CALLED IN FOR A LOCKDOWN"
		end
	end
	
	if reason then
		hud_text = hud_text .. ":" .. " " .. reason
	end
	
	managers.hud:show_hint({text = hud_text})
end

local function shouldGuardsSurrender()
	if level_id == "pex" then
		return false -- Police surrendering doesn't make sense
	end
	
	return true
end

-- Checks if a enemy is already assigned to group AI
function GroupAIStateBase:is_enemy_assigned(unit_key)
	return table.contains(self._police, unit_key)
end

function GroupAIStateBase:_tie_guards()
	if not shouldGuardsSurrender() then
		return
	end
	
	local is_security_pacified = self._has_tied_guards
	if is_security_pacified then
		return
	end
	
	self._has_tied_guards = true
	
	for u_key,u_data in pairs(managers.enemy:all_enemies()) do
		if alive(u_data.unit) and not u_data.unit:brain():surrendered() and not u_data.unit:brain():is_hostage() and not u_data.unit:unit_data().is_civilian then
			if u_data.unit:base():char_tweak().access == "security" or u_data.unit:base():char_tweak().access == "security_patrol" then
				u_data.unit:unit_data().is_civilian = true
				u_data.unit:brain():set_objective()
				u_data.unit:brain():set_logic("intimidated", nil)
			end
		end
	end
end

function GroupAIStateBase:_upd_teams()
	if level_id == "vit" then
		self:set_team_relation("neutral1", "criminal1", "foe", true)
		self:set_team_relation("neutral1", "converted_enemy", "foe", true)
	end
end

local function isTutorialHeist()
	if level_id == "short1_stage1" then
		return true
	elseif level_id == "short1_stage2" then
		return true
	elseif level_id == "short2_stage1" then
		return true
	elseif level_id == "short2_stage2b" then
		return true
	end
	
	return false
end

-- Pager guy will now raise alert level instead of instantly raising the alarm, triggering search mode.
-- This means you can answer the 5th pager.
local PAGER_TRIGGERS = {
	["alarm_pager_bluff_failed"] = true,
	["alarm_pager_hang_up"] = true,
	["alarm_pager_not_answered"] = true,
}

local isPagerGuySuspicious = false

local CAN_TRIGGER_SEARCH_MODE = {
	-- Cameras will still raise the alarm upon alert
	-- Since they cannot be unalerted and re-enabled again
	--[[
	["cam_dead_body"] = true,
	["cam_body_bag"] = true,
	["cam_criminal"] = true,
	["cam_criminal"] = true,
	["cam_hostage"] = true,
	["cam_drill"] = true,
	["cam_saw"] = true,
	["cam_sentry_gun"] = true,
	["cam_trip_mine"] = true,
	["cam_ecm_jammer"] = true,
	["cam_c4"] = true,
	["cam_computer"] = true,
	["cam_broken_cam"] = true,
	["cam_vault"] = true,
	["cam_fire"] = true,
	["cam_voting"] = true,
	["cam_glass"] = true,
	["cam_breaking_entering"] = true,
	["cam_crate_open"] = true,
	["cam_distress"] = true,
	]]
	["cop_criminal"] = true, -- spotted a criminal
	--["cop_gunfire"] = true, -- heard gunfire, commented out so they instantly raise alarm instead of searching
	["cop_dead_body"] = true, -- spotted a dead body
	["cop_body_bag"] = true, -- spotted a bag (or body bag)
	["cop_hostage"] = true, -- spotted a hostage
	["cop_drill"] = true, -- spotted a drill
	["cop_saw"] = true, -- spotted a saw
	["cop_sentry_gun"] = true, -- spotted a sentry gun
	["cop_trip_mine"] = true, -- spotted a trip mine (i don't think this works)
	["cop_ecm_jammer"] = true, -- spotted ecm jammer (ditto)
	["cop_c4"] = false, -- spotted c4 (prob explosion?)
	["cop_broken_cam"] = true, -- spotted broken security camera
	["cop_vault"] = true, -- spotted vault open
	["cop_fire"] = true, -- spotted fire?
	["cop_voting"] = true, -- spotted hack-in-progress voting machine
	["cop_glass"] = true, -- spotted broken glass
	["cop_breaking_entering"] = true, -- unsure what this does, i don't think they detect you trespassing
	["cop_crate_open"] = true, -- spotted a crate open
	["cop_computer"] = true, -- spotted hack-in-progress computer
	["cop_distress"] = true, -- spotted a civilian in distress
}

local on_police_called = Hooks:GetFunction(GroupAIStateBase, "on_police_called")
Hooks:OverrideFunction(GroupAIStateBase, "on_police_called", function(self, called_reason)
	-- Override the vanilla police_called event with our own
	-- So we can add things like search mode
	-- Search mode will only trigger on things like guard seeing suspicious activity or pager-related alarm calls, mission specific alarm triggers still raise the alarm normally
	
	if not self._ai_enabled then
		return
	end
	
	local was_called = self._police_called
	
	if was_called then
		return
	end
	
	if self._search_phase ~= "lockdown" and PAGER_TRIGGERS[called_reason] and not isTutorialHeist() then
		self:set_search_mode(true)
		return
	end
	
	if self._search_phase ~= "lockdown" and CAN_TRIGGER_SEARCH_MODE[called_reason] and not isTutorialHeist() then
		self:set_search_mode(true)
		return
	end
	
	self:_upd_teams()
	on_police_called(self, called_reason)
end)

Hooks:PostHook(GroupAIStateBase, "init", "pd3_groupai_init", function(self)
	self._is_search_mode_active = false
	self._search_phase = "calm"
	self._first_responders_arrived = false
end)

Hooks:PostHook(GroupAIStateBase, "_init_unit_type_filters", "add_new_type_filters", function(self)
	local nav_manager = managers.navigation
	local convert_f = nav_manager.convert_access_filter_to_number
	local guards_exclude_cops_filter = convert_f(nav_manager, {"security","security_patrol","murky","gangster"})
	self._unit_type_filter.guards_exclude_cops = guards_exclude_cops_filter
end)

Hooks:PostHook(GroupAIStateBase, "update", "pd3_groupai_update", function(self, t, dt)
	if not Network:is_server() then return end
	if not self:is_police_called() then
		if self:is_search_mode_active() then
			self:_upd_search_mode()
		end
	else
		if self._task_data.assault.active or self._first_responders_arrived then
			self:_tie_guards()
		end
	end
end)

local counter = 0

local function updateBotGivePlayerSupplies(self, t, dt)
	if self:team_ai_enabled() and Network:is_server() and Global.game_settings.single_player and managers.groupai:state():enemy_weapons_hot() and not self._task_data.assault.active then
		for _, ai in pairs(self:all_AI_criminals()) do
			local ai_pos = ai.unit:movement()._m_pos
			local closest_unit = nil
			local closest_distance = 200

			for _, player in pairs(self:all_player_criminals()) do
				local distance = mvector3.distance_sq(ai_pos, player.pos)

				if distance < closest_distance then
					closest_unit = player.unit
					closest_distance = distance
				end
			end

			if closest_unit then
				local max_health = closest_unit:character_damage():_max_health()
				local real_health = closest_unit:character_damage():get_real_health()
				
				if real_health < max_health then
					local pos = closest_unit:position()
					local rot = closest_unit:rotation()
					
					if math.random() < 0.15 then
						AmmoBagBase.spawn(pos, rot, 0)
					elseif math.random() < 0.25 then
						DoctorBagBase.spawn(pos, rot, 0)
					end
				end
			end
			
			break
		end
	end
end

Hooks:PostHook(GroupAIStateBase, "update", "botsGivePlayersSupplies", function(self, t, dt)
	if counter >= 1/30 then
		updateBotGivePlayerSupplies(self, t, dt)
		counter = 0
	else
		counter = counter + dt
	end
end)